<?php
namespace Ep\Gath\Plugin;

use Ep\Gath\DataStore\ProductDataStore;
use Ep\Gath\DataStore\VariableDataStore;
use Ep\Gath\DataStore\VariationDataStore;
use Ep\Gath\Gath;
use Ep\TecDoc\TecDocService;
use Solarium\Client;
use WC_Cart;
use WP_Post;
use WP_Query;
use WP_Term;
use Zend\Log\LoggerInterface;

/**
 * Zamjenjuje proizvode onima iz Gatha.
 *
 * @package Ep\Gath\Plugin
 * @author Elektronika Perhoč <web@ep.hr>
 * @link http://gitlab.ep.local/ep-framework/gath
 */
class Products extends AbstractGathPlugin
{
    const ID_OFFSET           = 1e7;
    const ID_OFFSET_SIMPLE    = self::ID_OFFSET;
    const ID_OFFSET_GROUPED   = self::ID_OFFSET_SIMPLE + 1e7;
    const ID_OFFSET_VARIABLE  = self::ID_OFFSET_GROUPED + 1e7;
    const ID_OFFSET_VARIATION = self::ID_OFFSET_VARIABLE + 1e7;

    const TERM_OFFSET       = 1e7;
    const TERM_OFFSET_BRAND = self::TERM_OFFSET;

    /** @var array */
    protected $resultsById = [];
    /** @var array */
    protected $termsById = [];

    /** @var Client */
    protected $search;
    /** @var TecDocService */
    protected $tecdoc;
    /** @var bool */
    protected $initialized = false;

    public function __construct(
        Gath $gath,
        LoggerInterface $logger,
        Client $search,
        TecDocService $tecdoc = null
    ) {
        parent::__construct($gath, $logger);
        $this->search = $search;
    }

    /**
     * Dodaje akcije i filtere u WordPress.
     *
     * @return void
     */
    public function __invoke(): void
    {
        if (!$this->initialized) {
            add_filter('get_order_item_metadata', [$this, 'get_order_item_metadata'], 10, 3);
            add_filter('get_post_metadata', [$this, 'get_post_metadata'], 10, 3);
            add_filter('get_term_metadata', [$this, 'get_term_metadata'], 10, 3);
            add_filter('get_the_terms', [$this, 'get_the_terms'], 10, 3);
            add_filter('posts_pre_query', [$this, 'posts_pre_query'], 10, 2);
            add_filter('update_post_metadata', [$this, 'update_post_metadata'], 10, 2);
            add_filter('woocommerce_add_to_cart_validation', [$this, 'wc_add_to_cart_validation'], 10, 2);
            add_action('init', [$this, 'preload_cart_products'], 20);
            add_action('woocommerce_restore_cart_item', [$this, 'wc_restore_cart_item'], 10, 2);

            add_action('wp_ajax_flatsome_quickview', [$this, 'flatsome_quickview'], 0);
            add_action('wp_ajax_nopriv_flatsome_quickview', [$this, 'flatsome_quickview'], 0);

            add_action('init', [$this, 'init']);
            add_action('plugins_loaded', [$this, 'plugins_loaded']);
            add_filter('woocommerce_redirect_single_search_result', '__return_false');
            add_filter('woocommerce_get_children', [$this, 'wc_get_children'], 10, 2);
            add_filter('woocommerce_product_type_query', [$this, 'wc_product_type_query'], 10, 2);
            add_filter('woocommerce_data_stores', [$this, 'wc_data_stores']);
            add_filter('woocommerce_structured_data_product', [$this, 'wc_structured_data_product'], 19);

            add_filter('ep_gath_products', [$this, 'ep_gath_products']);

            $this->initialized = true;
        }
    }

    public function plugins_loaded()
    {
        if (function_exists('ep_tecdoc_service')) {
            $this->tecdoc = ep_tecdoc_service();
        }
    }

    public function init()
    {
        remove_action('set_object_terms', ['WC_Post_Data', 'set_object_terms'], 10);
        remove_action('transition_post_status', ['WC_Post_Data', 'transition_post_status'], 10);
        remove_action('woocommerce_product_set_visibility', ['WC_Post_Data', 'delete_product_query_transients']);
    }

    /**
     * Ako imamo proizvoda u košarici, potrebno ih je preloadati prije nego što ih WooCommerce
     * pokuša učitati.
     */
    public function preload_cart_products()
    {
        $session = WC()->session;

        if (is_object($session) && method_exists($session, 'get')) {
            $cart = $session->get('cart');

            foreach ($cart ?? [] as $key => $value) {
                if (isset($value['product_id']) && $value['product_id'] > self::ID_OFFSET) {
                    $this->getProductsByIds([$value['product_id']]);
                }
            }
        }
    }

    /**
     * @param $metadata
     * @param $object_id
     * @param $meta_key
     *
     * @return mixed
     */
    public function get_order_item_metadata($metadata, $object_id, $meta_key)
    {
        if ('_product_id' === $meta_key) {
            global $wpdb;

            $productId = $wpdb->get_col("SELECT meta_value FROM wp_woocommerce_order_itemmeta WHERE meta_key = '_product_id' AND order_item_id = " . $object_id);

            if (count($productId) > 0) {
                $productId = intval($productId[0]);
                $this->getProductsByIds([$productId]);
            }
        }
        return $metadata;
    }

    /**
     * Ako ova metoda vrati postove, kveri se neće izvršiti.
     *
     * @param $null
     * @param WP_Query $query
     *
     * @return mixed
     *
     * @throws \Psr\Cache\InvalidArgumentException
     */
    public function posts_pre_query($null, WP_Query $query)
    {
        $data = null;

        if (!is_admin() && $query->is_main_query() && $query->is_search()) {
            return $this->product_search($null, $query);
        }

        // nije kveri za proizvode
        if (
            ($query->query_vars['wc_query'] ?? '') !== 'product_query' &&
            $query->query_vars['post_type'] !== 'product'
        ) {
            return $null;
        }

        // svima će trebat redoslijed
        switch ($query->query_vars['orderby'] ?? '') {
            case 'menu_order title':
                $sort = Gath::SORT_NAME_ASC;
                break;
            case 'price':
                $sort = $query->query_vars['order'] === 'DESC'
                    ? Gath::SORT_PRICE_DESC
                    : Gath::SORT_PRICE_ASC;
                break;
            case 'date ID':
            default:
                $sort = Gath::SORT_DATE_DESC;
                break;
        }

        // podaci o straničenju također
        $limit = $query->query_vars['posts_per_page'];
        $page = max(1, intval($query->query_vars['paged']));
        $offset = ($page - 1) * $limit;

        $tecdocKeys = $query->query_vars['tecdoc_keys'] ?? [];
        $motoGenIds = $query->query_vars['moto_genids'] ?? [];
        $motoId = absint($query->query_vars['moto_id'] ?? 0);

        $featuredTerm = get_term_by('slug', 'featured', 'product_visibility');

        if (isset($query->query['post__in']) && is_array($query->query['post__in']) && count($query->query['post__in']) > 0) {
            $ids = [];

            foreach ($query->query['post__in'] as $id) {
                if ($id > self::ID_OFFSET) {
                    $ids[] = $id;
                }
            }

            if (count($ids) > 0) {
                return $this->getProductsByIds($ids);
            }
        } elseif ('product' ===  ($query->query_vars['post_type'] ?? '') && isset($query->query_vars['product_id'])) {
            $productId = $query->query_vars['product_id'];

            return $this->getProductsByIds([$productId], $limit, $offset);
        // da li tražimo istaknute proizvode
        } elseif (
            isset($query->tax_query->queried_terms['product_visibility']) &&
            (in_array('featured', $query->tax_query->queried_terms['product_visibility']['terms']) ||
            in_array($featuredTerm->term_id, $query->tax_query->queried_terms['product_visibility']['terms']))
        ) {
            $data = $this->gath->getFeaturedProducts($sort, $limit, $offset);
        } elseif ($motoId > 0 && count($motoGenIds) > 0) {
            $data = $this->gath->getProductsByMotorcycle($motoId, $motoGenIds, $sort, $limit, $offset);
        // ako je ovo TecDoc kategorija
        } elseif (count($tecdocKeys) > 0) {
            $tecDocPairs = [];

            foreach ($tecdocKeys as $key) {
                [$brandNo, $articleNo] = explode('|', $key);

                $tecDocPairs[] = [
                    'BrandNo' => intval($brandNo),
                    'ArticleNo' => $articleNo,
                ];
            }

            $data = $this->gath->getProductsByTecDocPairs($tecDocPairs, $sort, $limit, $offset);
        } elseif (($query->query_vars['taxonomy'] ?? '') === 'product_cat') {
            $category = get_term_by('slug', $query->query_vars['term'], 'product_cat');
            $brands = isset($_GET['brands']) && !empty($_GET['brands']) ? array_filter(explode(',', $_GET['brands'])) : null;
            $data = $this->gath->getProductsByCategory($category->term_id, $sort, $limit, $offset, $brands);
            // proizvod u glavnom kveriju
        } elseif (preg_match('/-(\d{8})$/', $query->query_vars['name'], $matches)) {
            $productId = intval($matches[1]);
            return $this->getProductsByIds([$productId]);
        }
        
        
        if (isset($data) && count($data['Results']) > 0) {
            $query->set('no_found_rows', false);
            $query->found_posts = $data['RecordCount'];
            $query->max_num_pages = ceil($query->found_posts / $limit);
            $query->thumbnails_cached = true;

            return $this->loadPosts($data['Results']);
        }

        return null;
    }

    /**
     * @param $null
     * @param \WP_Query $query
     *
     * @return array
     *
     * @throws \Psr\Cache\InvalidArgumentException
     */
    public function product_search($null, \WP_Query $query)
    {
        if (is_admin() || !$query->is_main_query() || !$query->is_search()) {
            return $null;
        }

        $tecdoc_keys = $query->get('tecdoc_keys');
        $products = [];

        if (!is_array($tecdoc_keys)) {
            $tecdoc_keys = [];
        }

        $q = $query->get('s');

        if (strlen($q) > 0) {
            try {
                $queryString = sprintf('name:"%1$s" OR sku:"%2$s" OR attribute:"%1$s"', $q, strtoupper($q));
                $results = $this->getSearchResults($queryString);

                if ($results->count() === 0) {
                    $queryString = sprintf('name:%1$s~1 OR sku:%1$s OR attribute:%1$s', $q);
                    $results = $this->getSearchResults($queryString);
                }
            } catch (\Exception $e) {
                $results = [];
            }

            $productIds = [];
            $genericArticleIds = [];

            foreach ($results as $result) {
                if (in_array('product', $result->type)) {
                    if (is_array($result->parentid) && count($result->parentid) > 0) {
                        $productIds[] = self::ID_OFFSET_GROUPED + $result->parentid[0];
                    } else {
                        $productIds[] = absint($result->id);
                    }
                } else if (in_array('category', $result->type)) {
                    $multiIds = $result->tecdocmultiid[0] ?? '';

                    if (strlen($multiIds) > 0) {
                        $genericArticleIds = array_merge(
                            $genericArticleIds,
                            array_map('intval', explode(';', $result->tecdocmultiid[0] ?? ''))
                        );
                    }
                }
            }

            $productIds = array_unique($productIds);

            if (count($genericArticleIds) > 0 && $this->tecdoc) {
                $genericArticleIds = array_unique($genericArticleIds);
                $articles = $this->tecdoc->getArticlesByGenericIds($genericArticleIds);

                foreach ($articles as $article) {
                    $tecdoc_keys[] = $article['key'];
                }
            }

            if (count($productIds) > 0) {
                $products = array_merge($products, $this->getProductsByIds($productIds));
            }

            if (count($tecdoc_keys) > 0) {
                $tecdocProducts =  $this->getProductsByTecDocKeys($tecdoc_keys, 1, 1000, 0);
                $products = array_merge($products, $tecdocProducts);
            }
        }

        $query->set('no_found_rows', false);
        $query->found_posts = count($products);
        $query->max_num_pages = 1;
        $query->thumbnails_cached = true;
        $query->set('posts_per_page', -1);

        wc_set_loop_prop('total', count($products));
        return count($products) > 0 ? $products : $null;
    }

    /**
     * @param $productId
     *
     * @return array
     */
    public function getRawResult($productId): array
    {
        return $this->resultsById[$productId] ?? [];
    }

    /**
     * @param $queryString
     *
     * @return \Solarium\QueryType\Select\Result\Result
     */
    protected function getSearchResults($queryString)
    {
        $searchQuery = $this->search->createSelect();
        $searchQuery->setQuery($queryString);
        $searchQuery->setRows(10000);

        return $this->search->select($searchQuery);
    }

    /**
     * Vraća taksonomske pojmove vezane uz proizvod.
     *
     * @param $terms
     * @param $object_id
     * @param $taxonomy
     *
     * @return array
     */
    public function get_the_terms($terms, $object_id, $taxonomy)
    {
        if ($object_id > self::ID_OFFSET && isset($this->resultsById[$object_id])) {
            $result = $this->resultsById[$object_id];

            switch ($taxonomy) {
                case 'featured_item_category':
                case 'featured_item_tag':
                case 'product_cat':
                case 'product_tag':
                    return [];
                case 'product_brand':
                    if (isset($result['Brands']) && is_array($result['Brands']) && count($result['Brands']) === 1) {
                        return [$this->loadBrand($result['Brands'][0])];
                    } else if ($result['brandid'] > 0) {
                        $brand = get_term($result['brandid'], 'product_brand');
                        if ($brand instanceof WP_Term) {
                            return [$brand];
                        }
                    }
                    return [];
                case 'product_shipping_class':
                    return ($result['shippingclassid'] ?? 0) > 0
                        ? [get_term($result['shippingclassid'])]
                        : [];
                case 'product_type':
                    return [get_term_by('slug', ($result['type'] ?? 'simple'), 'product_type')];
                case 'product_visibility':
                    return $result['featured']
                        ? [get_term_by('slug', 'featured', 'product_visibility')]
                        : [];
            }
        }

        return $terms;
    }

    /**
     * @param $null
     * @param $object_id
     * @param $meta_key
     *
     * @return mixed
     */
    public function get_post_metadata($null, $object_id, $meta_key)
    {
        if ($object_id > self::ID_OFFSET && isset($this->resultsById[$object_id])) {
            $result = $this->resultsById[$object_id];
            $images = $result['Images'] ?? [];

            switch ($meta_key) {
                case '_automodel':
                    return $result['automodel'] ?? '';
                case '_automotor':
                    return $result['automotor'] ?? '';
                case '_backorders':
                case '_downloadable':
                case '_sold_individualy':
                case '_virtual':
                    return 'no';
                case '_children':
                    $childrenIds = [];
                    $children = $this->loadPosts($result['GroupedChildren'] ?? []);

                    foreach ($children as $child) {
                        $childrenIds[] = $child->ID;
                    }

                    return [$childrenIds];
                case '_product_image_gallery':
                    if (count($images) > 1) {
                        array_shift($images);
                        $urls = [];
                        foreach ($images as $image) {
                            $urls[] = 'gath-' . $image['url'];
                        }
                        return implode(',', $urls);
                    }
                    return '';
                case '_sku':
                    return $result['sku'] ?? '';
                case '_tax_class':
                    return $result['tax_class'] ?? ($result['tax_classHR'] ?? '');
                case '_tax_status':
                    return 'taxable';
                case '_tecdoc_key':
                case '_tecDoc_key':
                    return sprintf('%s|%s', $result['wsbrandno'], $result['wsartno']);
                case '_thumbnail_id':
                    return count($images) > 0
                        ? 'gath-' . $images[0]['url']
                        : $null;
                case '_crosssell_ids':
                case '_default_attributes':
                case '_download_expiry':
                case '_download_limit':
                case '_height':
                case '_length':
                case '_purchase_note':
                case '_sale_price_dates_from':
                case '_sale_price_dates_to':
                case '_upsell_ids':
                case '_weight':
                case '_width':
                case 'total_sales':
                    return '';
            }
        }

        return $null;
    }

    public function get_term_metadata($null, $object_id, $meta_key)
    {
        if (!isset($this->termsById[$object_id])) {
            return $null;
        }

        $term = $this->termsById[$object_id];

        switch ($meta_key) {
            case 'thumbnail_id':
                if (isset($term['image']) && !empty($term['image'])) {
                    return 'gath-' . $term['image'];
                }

                return '';
            default:
                return '';
        }
    }

    /**
     * Ne dopuštamo ažuriranje meta podataka za postove iz Gatha.
     *
     * @param $null
     * @param $object_id
     *
     * @return bool
     */
    public function update_post_metadata($null, $object_id)
    {
        if ($object_id > self::ID_OFFSET) {
            return true;
        }

        return $null;
    }

    /**
     * @return $this
     */
    public function ep_gath_products()
    {
        return $this;
    }

    /**
     * Prije nego što flatsome renda svoj quickview.
     */
    public function flatsome_quickview()
    {
        $this->getProductsByIds([intval($_POST['product'])]);
    }

    /**
     * Prije nego što dodajemo proizvod u košaricu.
     *
     * @param $valid
     * @param $product_id
     *
     * @return mixed
     */
    public function wc_add_to_cart_validation($valid, $product_id)
    {
        $this->getProductsByIds([$product_id]);
        return $valid;
    }

    /**
     * @param array $stores
     *
     * @return array
     */
    public function wc_data_stores(array $stores): array
    {
        $stores['product'] = ProductDataStore::class;
        $stores['product-variable'] = VariableDataStore::class;
        $stores['product-variation'] = VariationDataStore::class;
        return $stores;
    }

    /**
     * @param $children
     * @param \WC_Product $product
     *
     * @return array
     */
    public function wc_get_children($children, $product)
    {
        if (isset($this->resultsById[$product->get_id()])) {
            $result = $this->resultsById[$product->get_id()];
            $childrenIds = [];

            if (isset($result['GroupedChildren']) && is_array($result['GroupedChildren'])) {
                foreach ($result['GroupedChildren'] as $child) {
                    $childrenIds[] = $this->generateProductId($child);
                }
            }

            return $childrenIds;
        }
        return $children;
    }

    /**
     * @param $product_id
     *
     * @return string
     */
    public function wc_product_type_query($false, $product_id): string
    {
        if ($product_id > self::ID_OFFSET_VARIATION) {
            return 'variation';
        } else if ($product_id > self::ID_OFFSET_VARIABLE) {
            return 'variable';
        } else if ($product_id > self::ID_OFFSET_GROUPED) {
            return 'grouped';
        } else if ($product_id > self::ID_OFFSET_SIMPLE) {
            return 'simple';
        }

        return $false;
    }

    /**
     * Kada restoramo proizvod u košaricu, napunimo ga nazad proizvodom iz Gatha.
     *
     * @param string $cart_item_key
     * @param WC_Cart $cart
     */
    public function wc_restore_cart_item(string $cart_item_key, WC_Cart $cart)
    {
        $cartItem = $cart->cart_contents[$cart_item_key];

        $productId = absint($cartItem['variation_id'] ?: $cartItem['product_id']);

        if ($productId > 0) {
            $this->getProductById($productId);
            $cart->cart_contents[$cart_item_key]['data'] = wc_get_product($productId);
        }
    }

    public function wc_structured_data_product($markup)
    {

        return $markup;
    }

    /**
     * @param int $categoryId
     * @param int $orderBy
     * @param int $limit
     * @param int $offset
     *
     * @return array|WP_Post
     */
    public function getProductsByCategory(int $categoryId, int $orderBy = Gath::SORT_DATE_DESC, int $limit = 10, $offset = 0): array
    {
        $data = $this->gath->getProductsByCategory($categoryId, $orderBy, $limit, $offset);
        return $this->loadPosts($data['Results']);
    }

    /**
     * Generira ID koji koristimo unutar WordPress-a.
     *
     * @param array $product
     *
     * @return int
     */
    public function generateProductId(array $product): int
    {
        switch ($product['type'] ?? 'simple') {
            case 'variation':
                return self::ID_OFFSET_VARIATION + $product['id'];
            case 'variable':
                return self::ID_OFFSET_VARIABLE + $product['id'];
            case 'grouped':
                return self::ID_OFFSET_GROUPED + $product['id'];
            case 'simple':
            default:
                return self::ID_OFFSET_SIMPLE + $product['id'];
        }
    }

    /**
     * @param int $id
     *
     * @return WP_Post
     */
    public function getProductById(int $id): WP_Post
    {
        $products = $this->getProductsByIds([$id]);

        if (0 === count($products)) {
            throw new \RuntimeException(sprintf("Proizvod s id-em %d ne postoji u Gathu.", $id));
        }

        return $products[0];
    }

    /**
     * @param array $ids
     *
     * @return \WP_Post[]
     */
    public function getProductsByIds(array $ids, int $limit = 1000, int $offset = 0): array
    {
        $data = $this->gath->getProductsByIds($ids, $limit, $offset);
        return $this->loadPosts($data['Results']);
    }

    /**
     * @param array $skus
     *
     * @return \WP_Post[]
     */
    public function getProductsBySkus(array $skus): array
    {
        $data = $this->gath->getProductsByskus($skus);
        return $this->loadPosts($data['Results']);
    }

    /**
     * @param array $keys
     * @param int $orderBy
     * @param int $limit
     * @param int $offset
     *
     * @return \WP_Post[]
     */
    public function getProductsByTecDocKeys(array $keys, int $orderBy = Gath::SORT_DATE_DESC, int $limit = 10, $offset = 0): array
    {
        $tecDocPairs = [];

        foreach ($keys as $key) {
            [$brandNo, $articleNo] = explode('|', $key);

            $tecDocPairs[] = [
                'BrandNo' => intval($brandNo),
                'ArticleNo' => $articleNo,
            ];
        }

        return $this->getProductsByTecDocPairs($tecDocPairs, $orderBy, $limit, $offset);
    }

    /**
     * @param array $pairs
     * @param int $orderBy
     * @param int $limit
     * @param int $offset
     *
     * @return \WP_Post[]
     */
    public function getProductsByTecDocPairs(array $pairs, int $orderBy = Gath::SORT_DATE_DESC, int $limit = 10, $offset = 0): array
    {
        $data = $this->gath->getProductsByTecDocPairs($pairs, $orderBy, $limit, $offset);
        return $this->loadPosts($data['Results']);
    }

    /**
     * @param array $brand
     *
     * @return WP_Term
     */
    protected function loadBrand(array $brand): WP_Term
    {
        $brandId = $brand['brandid'] + self::TERM_OFFSET_BRAND;

        $this->termsById[$brandId] = $brand;

        return new WP_Term((object) [
            'term_id' => $brandId,
            'name' => $brand['name'],
        ]);
    }

    /**
     * @param array $results
     *
     * @return array|WP_Post[]
     */
    protected function loadPosts(array $results): array
    {
        $posts = [];

        foreach ($results as $result) {
            $postId = $this->generateProductId($result);
            $this->resultsById[$postId] = $result;

            $postDate = preg_replace('/\.\d+$/', '', $result['date_modified'] ?? date('Y-m-d H:i:s'));
            $postName = sprintf('%s-%d', sanitize_title($result['name'] ?? ''), $postId);

            $description = preg_replace('/<([a-z]+)[^>]+>/', '<\1>', $result['description'] ?? '');

            $post = get_post(new WP_Post((object) [
                'ID' => $postId,
                'post_author' => 1,
                'post_date' => $postDate,
                'post_date_gmt' => $postDate,
                'post_content' => $description,
                'post_title' => $result['name'] ?? '',
                'post_excerpt' => $this->gath->getShowShortDescription() ? ($result['short_description'] ?? '') : '',
                'post_status' => 'publish',
                'comment_status' => 'closed',
                'ping_status' => 'closed',
                'post_password' => '',
                'post_name' => $postName,
                'to_ping' => '',
                'pinged' => '',
                'post_modified' => $postDate,
                'post_modified_gmt' => $postDate,
                'post_content_filtered' => '',
                'post_parent' => 0,
                'guid' => sprintf('%s?p=%d', get_option('siteurl'), $postId),
                'post_type' => 'product',

                'filter' => 'raw',
            ]));

            wp_cache_add($postId, $post, 'posts');

            if (isset($result['GroupedChildren']) && is_array($result['GroupedChildren'])) {
                $this->loadPosts($result['GroupedChildren']);
            }

            $posts[] = $post;
        }

        global $wp_query;
        return apply_filters('the_posts', $posts, $wp_query);
    }
}